import os
from typing_extensions import Self
from ..template import Node
from .types import ModeFlags, ASCIISurface


__all__ = [
    "Frame",
    "Animation",
    "EmptyAnimation",
    "AnimationPlayer"
]


class Frame:
    def __init__(self, fpath: str) -> None:
        f = open(fpath)
        self.content = []
        for line in f.readlines():
            self.content.append(list((line.rstrip("\n"))))
        f.close()


class Animation:
    __slots__ = ("frames")
    def __init__(self, path: str, reverse: bool = False) -> None:
        fnames = os.listdir(path)
        step = 1 if not reverse else -1
        self.frames = [Frame(os.path.join(path, fname)) for fname in fnames][::step]


class EmptyAnimation(Animation):
    def __init__(self) -> None:
        self.frames = []


class AnimationPlayer(Node): # TODO: add buffered animations on load
    FIXED = 0 # TODO: implement FIXED and DELTATIME mode
    # DELTATIME = 1

    def __init__(self, owner: Self | None = None, fps: float = 16, /, mode: ModeFlags = FIXED, **animations) -> None:
        super().__init__(owner=None, x=0, y=0, z_index=0, force_sort=False) # TODO: change Node.owner --> Node.parent
        self.owner = owner
        self.fps = fps
        self.mode = mode # process mode (FIXED | DELTATIME)
        self.animations: dict[Animation] = dict(animations)
        self._animation = EmptyAnimation()
        self.is_playing = False
        self._current_frames = None
        self._next = None
        self._has_updated = False # indicates if the first frame (per animation) have been displayed
        self._accumulated_time = 0.0
        # -- dummy variables
        self.visible = False
        self.content = []
    
    def __iter__(self):
        return self

    def __next__(self) -> Frame:
        try:
            self._next = next(self._current_frames) # next of generator
            return self._next
        except StopIteration:
            self.is_playing = False
            self._current_frames = None
            self._next = None
            return None

    @property    
    def animation(self) -> Animation:
        return self._animation
    
    @animation.setter
    def animation(self, animation: str) -> None:
        self._animation = self.animations[animation]
        # make generator
        self._current_frames = (frame for frame in self._animation.frames)
        try:
            self._next = next(self._current_frames)
        except StopIteration:
            self.is_playing = False
            self._current_frames = None
            self._next = None
    
    def add(self, name: str, animation: Animation) -> None:
        self.animations[name] = animation
    
    def remove(self, name: str) -> None:
        del self.animations[name]
    
    def play(self, animation: str) -> None:
        """Plays an animation given the name of the animation

        Args:
            animation (str): the name of the animation to play
        """
        self.is_playing = True
        self._animation = animation
        self._current_frames = (frame for frame in self.animations[self.animation].frames)
        try:
            self._next: Frame = next(self._current_frames)
        except StopIteration:
            self.is_playing = False
            self._current_frames = None
            self._next: Frame = None
        if self._next != None:
            self.owner.content = self._next.content
            self._has_updated = False
    
    def play_backwards(self, animation: str) -> None:
        """Plays an animation backwards given the name of the animation

        Args:
            animation (str): the name of the animation to play backwards
        """
        self.is_playing = True
        self._animation = animation
        # reverse order frames
        self._current_frames = (frame for frame in reversed(self.animations[self.animation].frames))
        try:
            self._next: Frame = next(self._current_frames)
        except StopIteration:
            self.is_playing = False
            self._current_frames = None
            self._next: Frame = None
        if self._next != None:
            self.owner.content = self._next.content
            self._has_updated = False
        
    def advance(self) -> bool:
        """Advances 1 frame

        Returns:
            bool: whether it was NOT stopped
        """
        if self._current_frames == None:
            return False
        frame = self._next
        try:
            self._next = next(self._current_frames)
        except StopIteration:
            self.is_playing = False
            self._current_frames = None
            self._next = None
        if frame != None:
            self.owner.content = frame.content
            self._has_updated = False
        return frame != None # returns true if not stopped


    def stop(self) -> None:
        """Stops the animation from playing
        """
        self.is_playing = False

    def _render(self, surface: ASCIISurface) -> None: # dummy method
        return

    def _update(self, delta: float) -> None:
        if self.is_playing and self._has_updated:
            # if self.mode == AnimationPlayer.FIXED:
            frame = next(self)
            if frame == None:
                return
            self.owner.content = frame.content

            # elif self.mode == AnimationPlayer.DELTATIME:
            #     # apply delta time
            #     self._accumulated_time += delta
            #     if self._accumulated_time >= self._fps_ratio:
            #         self._accumulated_time -= self._fps_ratio # does not clear time
            #         frame = next(self)
            #         self.owner.content = frame.content
        self._has_updated = True



"""Template submodule from DisplayLib
"""

__all__ = [
    "Vec2",
    "overload",
    "OverloadUnmatched",
    "Node",
    "Engine",
    "Client"
]

from ..math import Vec2
from ..overload import overload, OverloadUnmatched
from .node import Node
from .engine import Engine
from .client import Client



from typing_extensions import Self
from ..node import ASCIINode


class ASCIILabel(ASCIINode):
    """Prefabricated `Label` node

    A new line is created for each `\\n`
    """
    def __init__(self, owner: Self | None = None, x: int = 0, y: int = 0, text: str = "", z_index: int = 0, force_sort: bool = True) -> None:
        super().__init__(owner, x, y, z_index, force_sort)
        self.text = text
    
    @property
    def text(self) -> str:
        """Returns a string from content

        Returns:
            str: content as string
        """
        return "\n".join("".join(line) for line in self.content)
    
    @text.setter
    def text(self, text: str) -> None:
        """Set content from string

        Args:
            text (str): string to be converted to content
        """
        self.content = [list(line) for line in text.split("\n")]



import pygame
from pygame.constants import *
from ..template import Node, Engine
from .constants import DEFAULT, MILLISECOND
from .types import Event, Surface


class PygameEngine(Engine):
    bg_color = (255, 255, 255) # white

    def __init__(self, window_name: str = "DisplayLib Window", tps: int = 60, width: int = 512, height: int = 256, icon_path: str | None = None, flags: int = DEFAULT) -> None:
        self._window_name = window_name # TODO: add setter
        self.tps = tps
        self._width = width # TODO: add setter
        self._height = height # TODO: add setter
        self.icon_img: None | Surface = None
        if icon_path:
            pygame.image.load(icon_path)
            pygame.display.set_icon(self.icon_img)
        self.flags = flags
        self.screen = pygame.display.set_mode(size=(width, height), flags=flags)
        self.display = pygame.Surface(size=(width, height))
        self._on_start()
        
        self.is_running = True
        self._main_loop()
    
    @property
    def height(self) -> int:
        return self._height
    
    @property
    def height(self, value: int) -> None: # TODO: queue this action
        self._height = value
        self.screen = pygame.display.set_mode(size=(self.width, self.height), flags=self.flags)
    
    @property
    def width(self) -> int:
        return self._width
    
    @property
    def width(self, value: int) -> None: # TODO: queue this action
        self._width = value
        self.screen = pygame.display.set_mode(size=(self.width, self.height), flags=self.flags)

    @property
    def icon(self) -> None:
        return self.icon_img
    
    @property
    def icon(self, icon_path: str) -> None:
        self.icon_img = pygame.image.load(icon_path)
        pygame.display.set_icon(self.icon_img)

    def _input(self, event: Event) -> None:
        if event.type == QUIT:
            self.is_running = False
    
    def _main_loop(self) -> None:
        def sort_fn(element):
            return element[1].z_index

        clock = pygame.time.Clock()
        delta = 1.0 / self.tps
        # update one time at the very start
        self.screen.fill(self.bg_color)
        pygame.display.flip()

        while self.is_running:
            self._update(delta)
            nodes = tuple(Node.nodes.values())
            for event in pygame.event.get():
                self._input(event)
                for node in nodes:
                    node._input(event)
            for node in nodes:
                node._update(delta)
            if Node._request_sort: # only sort once per frame if needed
                Node.nodes = {k: v for k, v in sorted(Node.nodes.items(), key=sort_fn)}
            
            self.display.fill(self.bg_color)
            for node in nodes: # render nodes onto the display
                node._render(self.display)
            # TODO: implement camera
            self.screen.blit(self.display, (0, 0)) # render display onto the main screen
            
            pygame.display.flip()
            delta = clock.tick(self.tps) / MILLISECOND # milliseconds -> seconds
        self._on_exit()



"""DisplayLib

Submodules:
- template
- ascii (default)
- pygame
"""

__version__ = "0.0.5"
__author__ = "FloatingInt"
__all__ = [
    "Vec2",
    "overload",
    "OverloadUnmatched",
    "ASCIINode",
    "ASCIIEngine",
    "ASCIICamera",
    "ASCIISurface",
    "ASCIIScreen",
    "ASCIIImage",
    "ASCIIClient",
    "Clock",
    ""
]

# utility
from .overload import overload, OverloadUnmatched
from .math import Vec2
# default module
from .ascii import (
    Node as ASCIINode,
    Engine as ASCIIEngine,
    Camera as ASCIICamera,
    Surface as ASCIISurface,
    Screen as ASCIIScreen,
    Image as ASCIIImage,
    Client as ASCIIClient,
    Frame as Frame,
    Animation as Animation,
    EmptyAnimation as EmptyAnimation,
    AnimationPlayer as AnimationPlayer,
    Clock as Clock,
    # -- prefab
    Label as ASCIILabel
)



from typing import Any
from types import FunctionType


class OverloadUnmatched(Exception):
    """Exception raised when no suitable overload is found based on the callers signature
    """
    def __init__(self, function_name: str, signature: str) -> None:
        super().__init__(self)
        self.function_name = function_name
        self.signature = signature
    
    def __str__(self) -> str:
        return f"function '{self.function_name}' has no overload of signature {self.signature}"


class overload:
    """Overload decorator

    Raises:
        OverloadUnmatched: no suitable overloaded function found

    Returns:
        function: overloaded function (also overwritten)
    """
    _uniques = {}

    def __init__(self, function: FunctionType) -> None:
        self.function_name = function.__name__
        hints = tuple(function.__annotations__.get(arg, object) for arg in function.__code__.co_varnames)
        signature = tuple(object if hint == Any else hint for hint in hints) # Any --> object
        if not function.__name__ in self._uniques:
            self._uniques[function.__name__] = {signature: function}
        else:
            self._uniques[function.__name__][signature] = function

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        signature = (*map(type, args), *map(type, kwargs.values()))
        print(signature)
        if signature in self._uniques[self.function_name]:
            function = self._uniques[self.function_name][signature]
            return function(*args, **kwargs)
        if (function := self._get_best_match(signature)):
            return function(*args, **kwargs)
        raise OverloadUnmatched(self.function_name, signature)

    def _get_best_match(self, signature: tuple) -> FunctionType | None:
        best_match = None
        best_points = 0

        for signature_option in self._uniques[self.function_name]:
            if len(signature) != len(signature_option):
                continue
            points = 0
            # print("OPTION:")
            for this, other in zip(signature, signature_option):
                # print(f"  CALLER ({this}) | POSSIBILITY ({other})")
                if issubclass(this, object):
                    points += 1
                elif this == other:
                    points += 1
            if points == best_points and points != 0: # prefer precise type over <object>
                # print("CLASHING:", best_match, "|", signature, "?", signature_option)
                a_points = 0
                b_points = 0
                for this, other, perfect in zip(best_match, signature, signature_option):
                    if perfect == object:
                        continue
                    if this != object:
                        a_points += 1
                    elif other != object:
                        b_points += 1
                best_match = signature_option if (a_points > b_points) else best_match
                # print("BEST MATCH:", best_match)
            elif points > best_points:
                best_points = points
                best_match = signature_option
        # print("BEST MATCH FINAL:", best_match)
        # assure 'best_match' isn't conflicting
        if best_match not in self._uniques[self.function_name]:
            return None
        return self._uniques[self.function_name][best_match]


if __name__ == "__main__":
    # Vec2 = tuple

    # @overload
    # def f(vec2: Vec2, num: float) -> Vec2:
    #     return (vec2[0] * num, vec2[1] * num)
    
    # @overload
    # def f(vec2: Vec2, other: Vec2) -> Vec2:
    #     return (vec2[0] + other[1])

    # print("- - -")
    # a = (2, 3)
    # b = (5, 7)
    # c = 2.0
    # print("A:", a)
    # print("A:", b)
    # print(f(a, b))
    # print(f(a, c))

    # @overload
    # def f(a: int, b: Any):
    #     print("B")
    #     return a + b

    # @overload
    # def f(a: int, b: int):
    #     print("A")
    #     return a + b
    
    # print(f(2, 3))

    @overload
    def f(a: int, b: int) -> int:
        print("A")
        return a + b
    
    @overload
    def f(a: str, b: str) -> str:
        print("B")
        return a + b
    
    @overload
    def f(a: Any, b: int) -> str:
        print("C")
        return a * b

    print(overload._uniques)
    print("- - -")
    print(f(2, 5))
    print(f([2], 3))
    print(f("2", "3"))



from ..template import Node
from .types import Event, Surface


class PygameNode(Node):
    visible: bool = False
    
    def _input(event: Event) -> None:
        return
    
    def _render(self, surface: Surface) -> None:
        return



import sys
from .node import ASCIINode
from .surface import ASCIISurface
from .constants import *


class ASCIIScreen(ASCIISurface):
    def show(self) -> None:
        lines = len(self.content)
        for idx, line in enumerate(self.content):
            rendered = "".join(letter if letter != ASCIINode.cell_transparant else ASCIINode.cell_default for letter in (line))
            sys.stdout.write(rendered + " " + ("\n" if idx != lines else ""))
        sys.stdout.write(ANSI_UP * len(self.content) + "\r")
        sys.stdout.flush()



"""ASCII submodule from DisplayLib
"""

__all__ = [
    "Vec2",
    "overload",
    "OverloadUnmatched",
    "Node",
    "Engine",
    "Camera",
    "Screen",
    "Surface",
    "Image",
    "Client",
    "Frame",
    "Animation",
    "EmptyAnimation",
    "AnimationPlayer",
    "Clock",
    # -- prefab
    "Label"
]

# -- standard imports
from ..math import Vec2
from ..overload import overload, OverloadUnmatched
# -- ascii imports
from .node import ASCIINode as Node
from .engine import ASCIIEngine as Engine
from .camera import ASCIICamera as Camera
from .surface import ASCIISurface as Surface
from .screen import ASCIIScreen as Screen
from .image import ASCIIImage as Image
from .client import ASCIIClient as Client
from .animation import Frame, Animation, EmptyAnimation, AnimationPlayer
from .clock import Clock
# -- prefab imports
from .prefab.label import ASCIILabel as Label


# -- activate ANSI escape codes
import os as _os
_os.system("")
# _os.system("cls") # used when developing
del _os



from typing_extensions import Self
from .node import Node


class Engine:
    """Engine base class
    NOTE: only one Engine instance should exist per script instance
    """
    tps: int = 4
    is_running: bool = False

    def __new__(cls: type[Self], *args, **kwargs) -> Self:
        instance = object.__new__(cls)
        Node.root = instance
        return instance

    def __init__(self) -> None: # default implementation
        self._on_start()
        self.is_running = True
        self._main_loop()
        self._on_exit()

    def _on_start(self) -> None:
        return
    
    def _on_exit(self) -> None:
        return
    
    def _update(self, delta: float) -> None:
        return
    
    def _main_loop(self) -> None:
        delta = 1.0 / self.tps
        while self.is_running:
            self._update(delta)
            for node in Node.nodes:
                node._update(delta)



from ..template import Node, Engine
from .clock import Clock
from .screen import ASCIIScreen
from .surface import ASCIISurface


class ASCIIEngine(Engine):
    """ASCIIEngine for creating a world in ASCII graphics
    """

    def __init__(self, tps: int = 16, width: int = 16, height: int = 8) -> None:
        self.tps = tps
        self.screen = ASCIIScreen(width=width, height=height) # upper level ASCIISurface
        self.display = ASCIISurface(Node.nodes.values()) # sub level ASCIISurface | just above a ASCIISurface
        self._on_start()

        self.is_running = True
        self._main_loop()
    
    def _render(self, surface: ASCIISurface) -> None:
        return
    
    def _main_loop(self) -> None:
        def sort_fn(element):
            return element[1].z_index

        nodes = tuple(Node.nodes.values())
        clock = Clock(self.tps)
        while self.is_running:
            delta = 1.0 / self.tps
            self.screen.clear()
            
            self._update(delta)
            for node in nodes:
                node._update(delta)

            for node in Node._queued_nodes:
                del Node.nodes[id(node)]
            Node._queued_nodes.clear()
            
            if Node._request_sort: # only sort once per frame if needed
                Node.nodes = {k: v for k, v in sorted(Node.nodes.items(), key=sort_fn)}
            nodes = tuple(Node.nodes.values())

            # render content of visible nodes onto a surface
            self.display = ASCIISurface(Node.nodes.values(), self.screen.width, self.screen.height) # create a Surface from all the Nodes
            self.screen.blit(self.display, transparent=True)
            
            self._render(self.screen)
            # render nodes onto main screen
            for node in nodes:
                node._render(self.screen)
            
            self.screen.show()
            clock.tick()
        
        # v exit protocol v
        self._on_exit()
        surface = ASCIISurface(Node.nodes.values(), self.screen.width, self.screen.height) # create a Surface from all the Nodes
        self.screen.blit(surface)
        self.screen.show()



import time
import socket
import selectors
from ..template import Node, Client
from .surface import ASCIISurface


class ASCIIClient(Client):
    _BUFF_SIZE = 4096
    _DELIMITER = b"$"
    _ARGUMENT_DELIMITER = ":"
    _NEGATIVE_INF = float("-inf")

    def __init__(self, host: str, port: int) -> None:
        self._address = (host, port)
        self._sel = selectors.DefaultSelector()
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sel.register(self._sock, selectors.EVENT_READ)
        self._sock.connect(self._address)
        self._sock.setblocking(False)
        self._buffer = bytes()

    def send(self, request: str) -> None:
        encoded = request.encode(encoding="utf-8") + self._DELIMITER
        self._sock.send(encoded)

    def _on_request(self, data: str) -> None:
        return
    
    def _update_socket(self) -> None:
        for key, mask in self._sel.select(timeout=self._NEGATIVE_INF):
            connection = key.fileobj
            if mask & selectors.EVENT_READ:
                data = connection.recv(self._BUFF_SIZE)
                if data: # a readable client socket that has data.
                    if self._DELIMITER in data:
                        head, *rest = data.split(self._DELIMITER)
                        self._buffer += head
                        data = self._buffer.decode()
                        request, *args = data.split(self._ARGUMENT_DELIMITER)
                        self._buffer = bytes()
                        self._on_request(request, list(args))
                        for content in rest[:-1]:
                            data = content.decode()
                            request, *args = data.split(self._ARGUMENT_DELIMITER)
                            self._on_request(request, list(args))
                        self._buffer += rest[-1]
                    else:
                        self._buffer += data
                    # print('  received {!r}'.format(data))
    
    def _main_loop(self) -> None:
        def sort_fn(element):
            return element[1].z_index

        while self.is_running:
            delta = 1.0 / self.tps
            if Node._request_sort: # only sort once per frame if needed
                Node.nodes = {k: v for k, v in sorted(Node.nodes.items(), key=sort_fn)}
            self._update_socket() # <--- updates socket
            self.screen.clear()
            self._update(delta)
            nodes = tuple(Node.nodes.values())
            for node in nodes:
                node._update(delta)
            # render nodes onto main screen
            surface = ASCIISurface(nodes, self.display.width, self.display.height) # create a Surface from all the Nodes
            self.screen.blit(surface)
            self.screen.display()
            
            time.sleep(delta) # TODO: implement clock
        self._on_exit()
        surface = ASCIISurface(nodes, self.display.width, self.display.height) # create a Surface from all the Nodes
        self.screen.blit(surface)
        self.screen.display()



"""Pygame submodule from DisplayLib

Raises:
    ModuleNotFoundError: `pygame` was not found
"""

__all__ = [
    "Vec2",
    "overload",
    "OverloadUnmatched",
    "Node",
    "Engine"
]

try: # check if pygame is installed
    import pygame as _pygame
    import os as _os
    _os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "True"
    del _os
    _pygame.init() # init without displaying message
    del _pygame
except ModuleNotFoundError as error:
    raise ModuleNotFoundError("pygame is required to import this submodule") from error

from ..math import Vec2
from ..overload import overload, OverloadUnmatched
from .node import PygameNode as Node
from .engine import PygameEngine as Engine



import io
import os
from ..math import Vec2
from .types import FilePath
from .surface import ASCIISurface


class _ASCIINodeStruct:
    def __init__(self, content: list[list]) -> None:
        self.visible = True
        self.content = content
        self.position = Vec2(0, 0)
        self.global_position = Vec2(0, 0)


class ASCIIImage:
    extension = ".txt"
    _cache = {}

    @classmethod
    def load(cls, file_path: FilePath, /, cache: bool = True) -> ASCIISurface:
        if not isinstance(file_path, str):
            TypeError("argument 'file_path' is required to be of type 'str'. '" + type(file_path).__name__ + "' found")
        fpath = os.path.normpath(file_path)
        if fpath in cls._cache and cache:
            return cls._cache[fpath]
        if not fpath.endswith(cls.extension):
            raise ValueError("argument 'file_path' needs to end with the current extension of '" + cls.extension + "'")
        file: io.TextIOWrapper = open(fpath, "r")
        content = list(map(list, file.readlines()))
        file.close()
        node = _ASCIINodeStruct(content)
        cls._cache[fpath] = node
        width = len(max(content, key=len))
        height = len(content)
        return ASCIISurface(nodes=[node], width=width, height=height)



from math import sqrt
from typing_extensions import Self


__all__ = ["Vec2"]


class Vec2:
    """Vector2 data structure

    Components: `x`, `y`

    Usefull for storing position or direction
    """
    def __init__(self, x: int | float = 0, y: int | float = 0) -> None:
        self.x = x
        self.y = y
    
    def __repr__(self) -> str:
        return f"Vec2({self.x}, {self.y})"
    
    def __add__(self, other: Self) -> Self:
        return Vec2(self.x + other.x, self.y + other.y)
    
    def __sub__(self, other: Self) -> Self:
        return Vec2(self.x - other.x, self.y - other.y)

    def __mul__(self, other: Self) -> Self:
        if isinstance(other, Vec2):
            return Vec2(self.x * other.x, self.y * other.y)
        elif isinstance(other, (int, float)):
            return Vec2(self.x * other, self.y * other)
    
    def __floordiv__(self, other: Self) -> Self:
        if isinstance(other, Vec2):
            return Vec2(self.x // other.x, self.y // other.y)
        elif isinstance(other, (int, float)):
            return Vec2(self.x // other, self.y // other)
    
    def __truediv__(self, other: Self) -> Self:
        if isinstance(other, Vec2):
            return Vec2(self.x / other.x, self.y / other.y)
        elif isinstance(other, (int, float)):
            return Vec2(self.x / other, self.y / other)
    
    def __mod__(self, other: int | float) -> Self:
        return Vec2(self.x % other, self.y % other)
    
    def __eq__(self, other: Self) -> bool:
        return (self.x == other.x) and (self.y == other.y)
    
    def __ne__(self, other: Self) -> bool:
        return (self.x != other.x) or (self.y != other.y)
    
    def __gt__(self, other: Self) -> bool:
        return self.x > other.x and self.y > other.y
    
    def __lt__(self, other: Self) -> bool:
        return self.x < other.x and self.y < other.y
    
    def __ge__(self, other: Self) -> bool:
        return self.x >= other.x and self.y >= other.y

    def __le__(self, other: Self) -> bool:
        return self.x <= other.x and self.y <= other.y

    def length(self) -> float:
        if self.x == 0 and self.y == 0:
            return 0.0
        return sqrt(self.x*self.x + self.y*self.y)
    
    def normalized(self) -> Self:
        return self / self.length()



from typing import Iterable
from typing_extensions import Self
from ..math import Vec2
from .node import ASCIINode
from .camera import ASCIICamera


class ASCIISurface:
    """ASCIISurface for displaying nodes
    """

    def __init__(self, nodes: Iterable[ASCIINode] = [], width: int = 16, height: int = 8) -> None:
        self._width = width
        self._height = height
        self.content = [[ASCIINode.cell_transparant for _ in range(width)] for _ in range(height)] # 2D array

        camera: ASCIICamera = ASCIICamera.current
        half_size = Vec2(self._width, self._height) // 2
        for node in nodes:
            if not node.visible:
                continue
            if not node.content:
                continue
            lines = len(node.content)
            longest = len(max(node.content, key=len))
            position = node.global_position - camera.global_position
            if camera.mode == ASCIICamera.CENTERED:
                position += half_size
            elif camera.mode == ASCIICamera.INCLUDE_SIZE:
                position -= Vec2(longest, lines) // 2
            elif camera.mode == ASCIICamera.CENTERED_AND_INCLUDE_SIZE:
                position += half_size
                position -= Vec2(longest, lines) // 2
            if position.y + lines < 0 or position.y > self._height:
                continue
            if position.x + longest < 0 or position.x > self._width:
                continue
            for h, line in enumerate(node.content):
                if not ((self._height) > (h + position.y) >= 0): # out of screen
                    continue
                for w, char in enumerate(line):
                    if not ((self._width) > (w + position.x) >= 0): # out of screen
                        continue
                    if char != ASCIINode.cell_transparant:
                        self.content[h+position.y][w+position.x] = char

    @property
    def width(self) -> int:
        return self._width
    
    @width.setter
    def width(self, value: int) -> None:
        self._width = value
        self.content = [[ASCIINode.cell_transparant for _ in range(self._width)] for _ in range(self._height)] # 2D array

    @property
    def height(self) -> int:
        return self._height
    
    @height.setter
    def height(self, value: int) -> None:
        self._height = value
        self.content = [[ASCIINode.cell_transparant for _ in range(self._width)] for _ in range(self._height)] # 2D array

    def clear(self) -> None:
        self.content = [[ASCIINode.cell_transparant for _ in range(self._width)] for _ in range(self._height)] # 2D array
    
    def blit(self, surface: Self, position: Vec2 = Vec2(0, 0), transparent: bool = False) -> None:
        lines = len(surface.content)
        longest = len(max(surface.content, key=len))
        if position.x > longest and position.y > lines: # completely out of screen
            return
        for h, line in enumerate(surface.content):
            if self._height < h + position.y or position.y < 0: # line out of screen
                continue
            for w, char in enumerate(line):
                if self._width < w + position.x or position.x < 0: # char out of screen
                    continue

                current = self.content[h+position.y][w+position.x]
                if current == ASCIINode.cell_default: # empty rendered cell
                    if not transparent:
                        self.content[h+position.y][w+position.x] = char
                        continue
                self.content[h+position.y][w+position.x] = char



import time


class Clock:
    def __init__(self, tps: int) -> None:
        self._tps = tps
        self._delta = 1.0 / self.tps
        self._last_tick = time.perf_counter()
    
    @property
    def tps(self) -> int:
        return self._tps
    
    @tps.setter
    def tps(self, value: int) -> None:
        self._tps = value
        self._delta = 1.0 / self._tps
    
    def tick(self):
        time.sleep(self._delta) # QUICKFIX
        return
        now = time.perf_counter()
        diff = now - self._last_tick
        self._last_tick = now
        remaining = self._delta - diff
        if remaining <= 0:
            return
        time.sleep(remaining)



from typing_extensions import Self
from ..math import Vec2
from .types import Engine


class Node:
    """Node base class

    Automatically keeps track of alive Node(s) by reference.
    An Engine subclass may access it's nodes through the `nodes` class attribute
    """
    root: Engine = None # set from a Engine subclass
    nodes = {} # all nodes that are alive
    _request_sort = False # requests Engine to sort
    _queued_nodes = set() # uses <Node>.queue_free() to ask Engine to delete them

    def __new__(cls: type[Self], *args, **kwargs) -> Self:
        instance = super().__new__(cls)
        Node.nodes[id(instance)] = instance
        return instance

    def __init__(self, owner: Self | None = None, x: int = 0, y: int = 0, z_index: int = 0, force_sort: bool = True) -> None:
        self.owner = owner
        self.position = Vec2(x, y)
        self._z_index = z_index
        self.visible = True
        if force_sort: # if True, requests sort every frame a new node is created
            Node._request_sort = True # otherwise, depent on a `z_index` change

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.position.x}, {self.position.y})"
    
    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def z_index(self) -> int:
        return self._z_index

    @z_index.setter
    def z_index(self, value: int) -> None:
        if self._z_index != value:
            self._z_index = value
            Node._request_sort = True
    
    @property
    def global_position(self) -> Vec2:
        position = self.position
        node = self.owner
        while node != None:
            position += node.position
            node = node.owner
        return position
    
    @global_position.setter
    def global_position(self, position: Vec2) -> None:
        diff = position - self.global_position
        self.position += diff

    def _update(self, delta: float) -> None:
        return
    
    def queue_free(self) -> None:
        Node._queued_nodes.add(self)



from setuptools import setup
from displaylib import __version__, __author__


setup(
   name="displaylib",
   version=__version__,
   description="Display an ASCII world",
   author=__author__,
   packages=[
      "displaylib",
      "displaylib.template",
      "displaylib.ascii",
      "displaylib.ascii.prefab",
      "displaylib.pygame"
   ]
)



from typing_extensions import Self
from ..math import Vec2
from ..template import Node
from .types import ASCIISurface


class ASCIINode(Node):
    cell_transparant = " " # type used to indicate that a cell is transparent in `content`
    cell_default = " " # the default look of an empty cell
    # per instance
    visible: bool = False

    def __init__(self, owner: Self | None = None, x: int = 0, y: int = 0, z_index: int = 0, force_sort: bool = True) -> None:
        super().__init__(owner, x, y, z_index)
        if force_sort:
            Node._request_sort = True # requests the Engine to sort every frame a new node is created
        self.content = [] # 2D array
    
    def _render(self, surface: ASCIISurface) -> None:
        return
    
    def _resize(self, size: Vec2) -> None:
        return



# ANSI codes
ANSI_UP = "\u001b[A"
ANSI_DOWN = "\u001b[B"
ANSI_RIGHT = "\u001b[C"
ANSI_LEFT = "\u001b[D"

# Camera flags
FIXED = 1
CENTERED = 3
INCLUDE_SIZE = 5
CENTERED_AND_INCLUDE_SIZE = 8

__all__ = [
    "ANSI_UP",
    "ANSI_DOWN",
    "ANSI_RIGHT",
    "ANSI_LEFT",
    "FIXED",
    "CENTERED",
    "INCLUDE_SIZE",
    "CENTERED_AND_INCLUDE_SIZE"
]



from typing_extensions import Self
from ..math import Vec2
from ..template import Node
from .node import ASCIINode
from .types import ASCIISurface, ModeFlags


class ASCIICamera(Node):
    """ASCIICamera for moving the viewport
    """
    FIXED = 1
    CENTERED = 3
    INCLUDE_SIZE = 5
    CENTERED_AND_INCLUDE_SIZE = 8
    current: Self

    def __init__(self, owner: ASCIINode | None = None, x: int = 0, y: int = 0, follow: bool = False, mode: ModeFlags = FIXED) -> None:
        self.owner = owner
        self.position = Vec2(x, y)
        self.mode = mode # `centered` mode only has effect if ´owner´ != None
        self.follow = follow # whether to follow the ´owner´
        self.content = [] # NOTE: attr ´content´ won't be used because ´visible´ will always be False
    
    @property
    def visible(self) -> int:
        return False # static, because it won't be displayed

    @property
    def z_index(self) -> int:
        return 0 # z_index is static, because it's required for the sort algorithm

    @property
    def z_index(self) -> int:
        return 0 # z_index static, because it's required for the sort algorithm
    
    def _update(self, delta: float) -> None:
        if self.follow and self.owner != None:
            self.global_position = self.owner.global_position
    
    def _render(self, surface: ASCIISurface) -> None:
        return

ASCIICamera.current = ASCIICamera() # default camera
